import os
import sys
import re

from .base_package import Service as BasePackageService

class Service(BasePackageService):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.sbt_credentials_path = os.getenv('SBT_CREDENTIALS_PATH', '%s/.sbt/.credentials' % os.environ['HOME'])

        self.name_regex = re.compile('name\s*:=\s*[\'\"]+(.*?)[\'\"]+')
        self.version_regex = re.compile('version(?: in ThisBuild)? := [\'\"]+(\d+\.\d+\.\d+\.*\d*)(?:-SNAPSHOT)?[\'\"]+')
        self.scala_version_regex = re.compile('(?:val )?scala(?:Version)?\s+:?= [\'\"]+(\d+\.\d+\.\d+)[\'\"]+')
        self.package_version_format = 'version in ThisBuild := "%s"'

        if 'package_file_path' in self.config['package_info'] and 'scala_version_file_path' not in self.config['package_info']:
             self.config['package_info']['scala_version_file_path'] = self.config['package_info']['package_file_path']

    def generate_sbt_credentials(self, credentials=None):
        if credentials:
            self.credentials = credentials
        else:   # setup defaults
            self.credentials = {
                'realm': os.environ['SBT_CREDENTIALS_REALM'],
                'host': os.environ['SBT_CREDENTIALS_HOST'],
                'user': os.environ['SBT_CREDENTIALS_USER'],
                'password': os.environ['SBT_CREDENTIALS_PASSWORD'],
            }

        print('generating credentials at %s' % self.sbt_credentials_path)
        os.makedirs(os.path.dirname(self.sbt_credentials_path), exist_ok=True)
        with open(self.sbt_credentials_path, 'w') as creds:
            for key, value in self.credentials.items():
                creds.write('%s=%s\n' % (key, value))

        print('credentials written to %s' % self.sbt_credentials_path)

    def read_package_info(self, package_info={}):
        super().read_package_info(package_info=package_info)

        self.package_info['scala_version'] = self.scala_version(self.config['package_info']['scala_version_file_path'])

    def scala_version(self, path):
        with open(path) as f:
            text = f.read()
            result = self.scala_version_regex.search(text)

            if result:
                return result.groups()[0]
            else:
                raise ValueError('Scala version not defined at: %s' % path)

    # only works with Artifactory :(
    def get_existing_versions(self, **package_info):
        import requests
        from requests.auth import HTTPBasicAuth

        minor_scala_version = '.'.join(package_info['scala_version'].split('.')[0:-1])
        package_path = '%s/%s_%s' % (package_info['organization'], package_info['name'], minor_scala_version)
        url = '%s/api/storage/%s/%s' % (os.environ['ARTIFACTORY_ENDPOINT'], package_info['repository'], package_path)

        response = requests.get(
            url,
            auth=HTTPBasicAuth(os.environ['ARTIFACTORY_USERNAME'], os.environ['ARTIFACTORY_PASSWORD']),
        )

        versions = []
        if response.status_code == 404:
            print('WARNING: repository path not found: %s' % url)
        else:
            response.raise_for_status()

            body = response.json()
            versions.extend([child['uri'][1:] for child in body['children'] if child['folder']])

        print('existing versions for: %s' % package_info['name'])
        print(versions)

        return versions



import os

class BaseCIAdapter():

    def __init__(self):
        self.env_map = {}

    def convert_environment_variables(self):
        for key, value in os.environ.items():
            if key in self.env_map and not self.env_map[key] in os.environ:
                os.environ[self.env_map[key]] = value
                os.environ.pop(key)



from unittest import TestCase, mock
import os
from datetime import datetime

from servicer.servicer import Servicer

class ServicerTest(TestCase):
    def setUp(self):
        self.servicer = Servicer(args={})

    def AutoMock(self, **attributes):
        automock = mock.Mock(spec=attributes.keys())
        for k, v in attributes.items():
            setattr(automock, k, v)
        return automock

class ServicerClassTest(ServicerTest):
    def test_initialized(self):
        pass

class LoadEnvironmentTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.datetime = self.AutoMock(
            now=mock.Mock(return_value=self.AutoMock(strftime=mock.Mock(return_value='1970-01-01'))),
            utcnow=mock.Mock(return_value='1970-01-01 00:00:00'),
        )
        self.servicer.load_env_file = mock.Mock(return_value=None)

        self.args = {}

    def test_environment_variables_set(self):
        os.environ['PWD'] = 'project_path'
        result = self.servicer.load_environment(self.args)
        self.assertEqual(os.environ['PROJECT_PATH'], 'project_path')
        self.assertEqual(os.environ['BUILD_DATETIME'], '1970-01-01 00:00:00')
        self.assertEqual(os.environ['BUILD_DATE'], '1970-01-01')

    def test_no_env_file_paths(self):
        result = self.servicer.load_environment(self.args)
        self.servicer.load_env_file.assert_not_called()

    def test_multiple_env_file_paths(self):
        self.args['env_file_paths'] = 'path/one:path/two'
        result = self.servicer.load_environment(self.args)
        print(self.servicer.load_env_file.mock.calls)
        self.servicer.load_env_file.assert_has_calls([
            mock.call('path/one'),
            mock.call('path/two'),
        ])

class LoadEnvFileTest(ServicerTest):
    def setUp(self):
        super().setUp()

    def test_path_does_not_exist(self):
        with mock.patch('builtins.open', create=True) as mock_open:
            result = self.servicer.load_env_file('fake/path.yaml')
            mock_open.assert_not_called()

    def test_valid_env_file_path(self):
        with mock.patch('yaml.load', create=True) as mock_yaml_load:
            with mock.patch('builtins.open', create=True) as mock_open:
                with mock.patch('os.path.exists', create=True) as mock_os_path_exists:
                    mock_yaml_load.return_value = {'FOO': 'BAR'}
                    mock_os_path_exists.return_value = True

                    result = self.servicer.load_env_file('real/path.yaml')
                    self.assertEqual(os.environ['FOO'], 'BAR')

class GetServiceEnvironmentTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.map_service_environment = mock.Mock(return_value=None)

        self.servicer.config['environment'] = {
            'mappings': [{ 'branch': 'foo', 'environment': 'bar' }],
        }

    def test_custom_mappings(self):
        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', self.servicer.config['environment']['mappings'])
        self.assertEqual(result, None)

    def test_matched_environment(self):
        self.servicer.map_service_environment.return_value = 'my_wacky/environment'

        result = self.servicer.get_service_environment('my-branch')
        self.servicer.map_service_environment.assert_called_with('my-branch', mock.ANY)
        self.assertEqual(result, 'my-wacky-environment')

class MapServiceEnvironmentTest(ServicerTest):
    def test_no_mappings(self):
        result = self.servicer.map_service_environment('my-branch', [])
        self.assertEqual(result, None)

    def test_branch_mapping(self):
        result = self.servicer.map_service_environment('develop', [
            { 'branch': 'develop' },
        ])
        self.assertEqual(result, 'develop')

    def test_branch_environment_mapping(self):
        result = self.servicer.map_service_environment('master', [
            { 'branch': 'master', 'environment': 'production' },
        ])
        self.assertEqual(result, 'production')

    def test_branch_wildcard_mapping(self):
        result = self.servicer.map_service_environment('env-my-branch', [
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, 'env-my-branch')

    def test_regex_mapping(self):
        result = self.servicer.map_service_environment('env-my-branch', [
            { 'branch': '/.*-my-.*/' },
        ])
        self.assertEqual(result, 'env-my-branch')

    def test_no_match(self):
        result = self.servicer.map_service_environment('my-branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, None)

    def test_no_match_but_starts_with_branch(self):
        result = self.servicer.map_service_environment('master_branch', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': 'develop' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, None)

    def test_multiple_match(self):
        result = self.servicer.map_service_environment('env-my-branch-qa', [
            { 'branch': 'master', 'environment': 'production' },
            { 'branch': '*-qa', 'environment': 'qa' },
            { 'branch': 'env-*' },
        ])
        self.assertEqual(result, 'qa')

class OrderServiceStepsTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.calculate_service_dependencies = mock.Mock(side_effect=[
            {'service_1:test': set(['service_1:build'])},
            {'service_2:test': set(['service_2:build'])},
        ])
        self.servicer.toposort2 = mock.Mock(return_value=[
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ])
        self.servicer.config = {
            'services': {
                'service_1': {
                    'field1': 'value1',
                },
                'service_2': {
                    'field2': 'value2',
                },
            },
        }

    def test_returns_a_service_step_order(self):
        result = self.servicer.order_service_steps(self.servicer.config['services'].keys())

        self.servicer.calculate_service_dependencies.assert_has_calls([
            mock.call(self.servicer.config['services']['service_1']),
            mock.call(self.servicer.config['services']['service_2']),
        ])
        self.servicer.toposort2.assert_called_with({
            'service_1:test': set(['service_1:build']),
            'service_2:test': set(['service_2:build']),
        })
        self.assertEqual(result, [
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ])

    def test_returns_an_empty_service_step_order(self):
        self.servicer.toposort2 = mock.Mock(return_value=[])

        result = self.servicer.order_service_steps([])

        self.servicer.calculate_service_dependencies.assert_not_called()
        self.servicer.toposort2.assert_called_with({})
        self.assertEqual(result, [])

class CalculateServiceDependenciesTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.add_dependencies = mock.Mock()

        self.servicer.step_order = [
            'build',
            'test',
            'deploy',
        ]

        self.service = {
            'name': 'service_1',
            'steps': {
                'build': {'config': {}},
                'test': {'config': {}},
                'deploy': {'config': {}},
            },
        }

    def test_calculates_standard_step_order(self):
        result = self.servicer.calculate_service_dependencies(self.service)

        self.servicer.add_dependencies.assert_has_calls([
            mock.call({}, self.service, 'test', 'service_1:build', persistent_steps=True),
            mock.call({}, self.service, 'deploy', 'service_1:test', persistent_steps=True),
        ])
        self.assertEqual(result, {})

    def test_ignores_steps_that_does_not_exist(self):
        self.service['steps']['bad_step'] = {'config': {}}

        result = self.servicer.calculate_service_dependencies(self.service)

        self.servicer.add_dependencies.assert_has_calls([
            mock.call({}, self.service, 'test', 'service_1:build', persistent_steps=True),
            mock.call({}, self.service, 'deploy', 'service_1:test', persistent_steps=True),
        ])
        self.assertEqual(result, {})

    def test_calculates_custom_service_dependencies(self):
        self.service['depends_on'] = [
            'service_2',
        ]

        result = self.servicer.calculate_service_dependencies(self.service)

        self.servicer.add_dependencies.assert_has_calls([
            mock.call({}, self.service, 'build', ['service_2']),
            mock.call({}, self.service, 'test', 'service_1:build', persistent_steps=True),
            mock.call({}, self.service, 'test', ['service_2']),
            mock.call({}, self.service, 'deploy', 'service_1:test', persistent_steps=True),
            mock.call({}, self.service, 'deploy', ['service_2']),
        ])
        self.assertEqual(result, {})

    def test_calculates_custom_step_dependencies(self):
        self.service['steps']['test']['depends_on'] = [
            'service_2',
            'service_3:build',
        ]

        result = self.servicer.calculate_service_dependencies(self.service)

        self.servicer.add_dependencies.assert_has_calls([
            mock.call({}, self.service, 'test', 'service_1:build', persistent_steps=True),
            mock.call({}, self.service, 'test', ['service_2', 'service_3:build']),
            mock.call({}, self.service, 'deploy', 'service_1:test', persistent_steps=True),
        ])
        self.assertEqual(result, {})

class AddDependenciesTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.config = {
            'services': {
                'service_1': {
                    'name': 'service_1',
                    'steps': {'build': {}, 'test': {}},
                },
                'service_2': {
                    'name': 'service_2',
                    'steps': {'build': {}, 'test': {}, 'deploy': {}},
                },
                'service_3': {
                    'name': 'service_3',
                    'steps': {'build': {}, 'test': {}},
                },
                'service_4': {
                    'name': 'service_4',
                    'steps': {'build': {}, 'deploy': {}},
                },
            },
        }

        self.servicer.steps = {
            'build': {'name': 'build'},
            'test': {'name': 'test'},
            'deploy': {
                'name': 'deploy',
                'config': {'persists': True},
            },
        }

        self.servicer.active_services = self.servicer.config['services'].keys()

        self.service = {
            'name': 'service_1',
            'steps': {
                'build': {'config': {}},
                'test': {'config': {}},
                'deploy': {'config': {}},
            },
        }
        self.dependencies = {}

    def test_adds_a_single_service_step_dependency(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'build', 'service_2:build')

        self.assertEqual(self.dependencies, {
            'service_1:build': set(['service_2:build']),
        })

    def test_adds_a_list_of_service_dependencies(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'test', ['service_2', 'service_3'])

        self.assertEqual(self.dependencies, {
            'service_1:test': set(['service_2:test', 'service_3:test']),
        })

    def test_skips_adding_nonexistant_steps_for_a_service_dependency(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'test', 'service_4')

        self.assertEqual(self.dependencies, {
            'service_1:test': set(),
        })

    def test_adding_a_service_name_wildcard(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'test', ['*:test'])

        self.assertEqual(self.dependencies, {
            'service_1:test': set(['service_1:test', 'service_2:test', 'service_3:test']),
        })

    def test_adds_a_list_of_service_step_dependencies(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'deploy', [
            'service_2:build',
            'service_3:test',
        ])

        self.assertEqual(self.dependencies, {
            'service_1:deploy': set(['service_2:build', 'service_3:test']),
        })

    def test_skips_persists_steps_by_default(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'deploy', [
            'service_2',
            'service_3:deploy',
        ])

        self.assertEqual(self.dependencies, {
            'service_1:deploy': set(),
        })

    def test_does_not_skips_a_persists_step(self):
        result = self.servicer.add_dependencies(self.dependencies, self.service, 'deploy', [
            'service_2',
            'service_3:deploy',
        ], persistent_steps=True)

        self.assertEqual(self.dependencies, {
            'service_1:deploy': set(['service_2:deploy', 'service_3:deploy']),
        })

    def test_errors_if_service_does_not_exist(self):
        with self.assertRaises(ValueError) as context:
            result = self.servicer.add_dependencies(self.dependencies, self.service, 'test', 'service_0')

        self.assertTrue('Invalid service dependency specified: service_0, "service_0" must be included in services: [service_1,service_2,service_3,service_4]' in str(context.exception))

    def test_errors_if_step_does_not_exist(self):
        with self.assertRaises(ValueError) as context:
            result = self.servicer.add_dependencies(self.dependencies, self.service, 'build', 'service_2:bad_step')

        self.assertTrue('Invalid step dependency specified: service_2:bad_step, "bad_step" must be included in steps: [build,test,deploy]' in str(context.exception))

class RunServiceStepsTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.run_service_step = mock.Mock()

        self.servicer.steps = {
            'build': {'name': 'build'},
            'test': {'name': 'test'},
            'deploy': {'name': 'deploy', 'config': {'requires_service_environment': True}},
        }

        self.servicer.config = {
            'services': {
                'service_1': {
                    'name': 'service_1',
                    'steps': {
                        'build': {'config': {}},
                        'test': {'config': {}},
                        'deploy': {'config': {'type': 'deploy'}},
                    },
                },
                'service_2': {
                    'name': 'service_2',
                    'steps': {
                        'build': {'config': {}},
                        'test': {'config': {}},
                        'deploy': {'config': {'type': 'deploy'}},
                    },
                },
            },
        }

        self.servicer.service_step_order = [
            ['service_1:build', 'service_2:build'],
            ['service_1:test', 'service_2:test'],
        ]

    def test_runs_all_service_steps(self):
        result = self.servicer.run_service_steps()

        self.servicer.run_service_step.assert_has_calls([
            mock.call(self.servicer.config['services']['service_1'], {'config': {}}),
            mock.call(self.servicer.config['services']['service_2'], {'config': {}}),
            mock.call(self.servicer.config['services']['service_1'], {'config': {}}),
            mock.call(self.servicer.config['services']['service_2'], {'config': {}}),
        ])

    def test_skips_service_steps_that_require_a_service_environment(self):
        self.servicer.service_step_order = [
            ['service_1:build'],
            ['service_1:deploy'],
        ]

        result = self.servicer.run_service_steps()

        self.servicer.run_service_step.assert_has_calls([
            mock.call(self.servicer.config['services']['service_1'], {'config': {}}),
        ])

    def test_runs_service_steps_that_require_a_service_environment(self):
        self.servicer.service_environment = 'duck_tales'
        self.servicer.service_step_order = [
            ['service_1:build'],
            ['service_1:deploy'],
        ]

        result = self.servicer.run_service_steps()

        self.servicer.run_service_step.assert_has_calls([
            mock.call(self.servicer.config['services']['service_1'], {'config': {}}),
            mock.call(self.servicer.config['services']['service_1'], {'config': {'type': 'deploy'}}),
        ])

class RunServiceStepTest(ServicerTest):
    def setUp(self):
        super().setUp()

        self.servicer.run = mock.Mock()
        self.servicer.interpolate_tokens = mock.Mock()

        self.servicer.config = {}

        self.adapter = mock.Mock()
        self.adapter.up = mock.Mock(return_value='service-step results')
        self.Service = mock.Mock(return_value=self.adapter)
        self.module = self.AutoMock(Service=self.Service)

        def mock_load_service_module(service):
            service['module'] = self.module

        self.servicer.load_service_module = mock.Mock(side_effect=mock_load_service_module)

        self.service = {
            'module': self.module,
            'name': 'service_1',
            'steps': {
                'build': {'config': {}},
            }
        }

    def test_runs_an_empty_service_step(self):
        self.service.pop('module')
        self.service['steps']['build'].pop('config')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.interpolate_tokens.assert_not_called()

    def test_runs_a_service_step_with_config_and_no_module(self):
        self.service.pop('module')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.load_service_module.assert_called_with(self.service)

        self.servicer.run.assert_not_called()
        self.servicer.interpolate_tokens.assert_called_with(self.service['steps']['build']['config'], self.servicer.config, ignore_missing_key=True)
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_module_and_no_config(self):
        self.service['steps']['build'].pop('config')

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.interpolate_tokens.assert_not_called()

    def test_runs_a_service_step_with_module_and_config(self):
        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.interpolate_tokens.assert_called_with(self.service['steps']['build']['config'], self.servicer.config, ignore_missing_key=True)

        self.assertTrue('git' not in self.service['steps']['build']['config'])
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_git_integration(self):
        self.servicer.config['git'] = {'enabled': True}
        self.servicer.git = {}

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_not_called()
        self.servicer.interpolate_tokens.assert_called_with(self.service['steps']['build']['config'], self.servicer.config, ignore_missing_key=True)

        self.assertEqual(self.service['steps']['build']['config']['git'], {'module': {}})
        self.assertEqual(self.service['steps']['build']['results'], 'service-step results')

    def test_runs_a_service_step_with_commands(self):
        self.service['steps']['build']['commands'] = [
            'pre-command1.sh',
            'rm -rf treeeeee xD',
        ]
        self.service['steps']['build']['post_commands'] = [
            'cowsay moo',
            'yes | lolcat',
        ]

        result = self.servicer.run_service_step(self.service, self.service['steps']['build'])

        self.servicer.run.assert_has_calls([
            mock.call('pre-command1.sh'),
            mock.call('rm -rf treeeeee xD'),
            mock.call('cowsay moo'),
            mock.call('yes | lolcat'),
        ])



from unittest import TestCase, mock

from servicer.config_loader import ConfigLoader

class ConfigLoaderTest(TestCase):
    def setUp(self):
        self.config_loader = ConfigLoader(args={})

class ConfigLoaderClassTest(ConfigLoaderTest):
    def test_initialized(self):
        pass

class MergeConfigTest(ConfigLoaderTest):
    def setUp(self):
        super().setUp()

        self.config_loader.merge_config = mock.Mock(side_effect=self.config_loader.merge_config)

    def test_merges_empty_dicts(self):
        from_dict = {}
        to_dict = {}

        self.config_loader.merge_config(to_dict, from_dict)

        self.config_loader.merge_config.assert_has_calls([
            mock.call(to_dict, from_dict),
        ])
        self.assertEqual(to_dict, {})

    def test_merges_empty_dict_into_non_empty(self):
        from_dict = {}
        to_dict = { 'foo': 'bar' }

        self.config_loader.merge_config(to_dict, from_dict)

        self.config_loader.merge_config.assert_has_calls([
            mock.call(to_dict, from_dict),
        ])
        self.assertEqual(to_dict, { 'foo': 'bar' })

    def test_overwrites_values(self):
        from_dict = {
            'tacos': 'tacodeli',
        }
        to_dict = {
            'foo': 'bar',
            'tacos': 'torchys',
        }

        self.config_loader.merge_config(to_dict, from_dict)

        self.config_loader.merge_config.assert_has_calls([
            mock.call(to_dict, from_dict),
        ])
        self.assertEqual(to_dict, {
            'foo': 'bar',
            'tacos': 'tacodeli',
        })

    # note: list values are not combined, they are completely overwritten
    def test_overwrites_nested_values(self):
        from_dict = {
            'numbers': [4, 5, 6],
            'people': {
                'astronaut': 'buzz aldrin',
            },
            'capitols': [
                { 'brazil': 'not rio de janeiro' },
            ],
        }
        to_dict = {
            'numbers': [1, 2, 3],
            'people': {
                'astronaut': 'neil armstrong',
                'philosopher': 'socrates',
            },
            'capitols': [
                { 'france': 'paris' },
                { 'brazil': 'rio de janeiro' },
            ],
        }

        self.config_loader.merge_config(to_dict, from_dict)

        self.config_loader.merge_config.assert_has_calls([
            mock.call(to_dict, from_dict),
        ])
        self.assertEqual(to_dict, {
            'numbers': [4, 5, 6],
            'people': {
                'astronaut': 'buzz aldrin',
                'philosopher': 'socrates',
            },
            'capitols': [
                { 'brazil': 'not rio de janeiro' },
            ],
        })



import re

def interpolate_tokens(config, params, ignore_missing_key=False):
    if isinstance(config, dict):
        for key, value in config.items():
            if isinstance(value, str):
                config[key] = replace_tokens(value, params, ignore_missing_key)
            else:
                interpolate_tokens(value, params, ignore_missing_key)
    elif isinstance(config, list):
        for i in range(len(config)):
            if isinstance(config[i], str):
                config[i] = replace_tokens(config[i], params, ignore_missing_key)
            else:
                interpolate_tokens(config[i], params, ignore_missing_key)

def replace_tokens(value, params, ignore_missing_key=False):
    escaped_values = [
        '*',
    ]

    for match in re.findall(r'\${.+?}+', value):
        token = match[2:-1]
        replace_value = None

        pieces = token.split(':')
        pieces = [evaluate_value(p, params) for p in pieces]
        key = pieces[0]

        if key not in params:
            if ignore_missing_key:
                replace_value = '${%s}' % ':'.join(pieces)
            elif len(pieces) > 1:
                # use default value
                replace_value = pieces[1]

        if replace_value == None:
            replace_value = params[key]

        if replace_value != None:
            for ev in escaped_values:
                if ev in token:
                    token = token.replace(ev, '\%s' % ev)

            value = re.sub(r'\${%s}' % token, replace_value, value)

    return value

def evaluate_value(value, params):
    if not (value.startswith('{') and value.endswith('}')):
        return value

    path_pieces = value[1:-1].split('.')
    if len(path_pieces) > 0:
        path_value = dict_get_path(_dict=params, path=path_pieces, ignore_missing_key=True)
        if path_value:
            value = path_value

    return value

def dict_get_path(_dict={}, path=[], ignore_missing_key=False):
    if path == None or len(path) < 1:
        return None

    current_path = path.pop(0)
    if ignore_missing_key and current_path not in _dict:
        return None

    if len(path) < 1:
        return _dict[current_path]

    return dict_get_path(_dict=_dict[current_path], path=path)



import os
import sys
import re

from .base_package import Service as BasePackageService

class Service(BasePackageService):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.pypirc_path = os.getenv('PYPIRC_PATH', '%s/.pypirc' % os.environ['HOME'])
        self.pip_conf_path = os.getenv('PIP_CONF_PATH', '%s/.pip/pip.conf' % os.environ['HOME'])

        self.name_regex = re.compile('name\s*=\s*[\'\"]+(.*?)[\'\"]+')
        self.version_regex = re.compile('version\s*=\s*[\'\"]+(\d+\.\d+\.\d+)[\'\"]+')
        self.package_version_format = 'version=\'%s\''

    def setup_py(self, command):
        self.run('%s setup.py %s' % (os.getenv('PYTHON_EXE', 'python'), command))

    def generate_pypi_config(self, config=None):
        if config:
            self.pypi_config = config
        else:   # setup defaults
            self.pypi_config = {
                'pypirc': {
                    'servers': {
                        'testpypi': {
                            'repository': 'https://test.pypi.org/legacy/',
                            'username': os.environ['PYPI_USERNAME'],
                            'password': os.environ['PYPI_PASSWORD'],
                        },
                        'pypi': {
                            'username': os.environ['PYPI_USERNAME'],
                            'password': os.environ['PYPI_PASSWORD'],
                        },
                    },
                },
            }

            if os.getenv('PYPI_REPOSITORY'):
                pieces = os.environ['PYPI_REPOSITORY'].split('://')
                protocol = pieces[0]
                domain = pieces[1]

                url = '%s://%s:%s@%s' % (protocol, os.environ['PYPI_USERNAME'], os.environ['PYPI_PASSWORD'], domain)
                self.pypi_config['pip.conf'] = {
                    'entries': [{
                        'scope': 'global',
                        'args': {
                            'extra-index-url': url,
                        },
                    }],
                }
        if 'pypirc' in self.pypi_config:
            self.generate_pypirc()
        if 'pip.conf' in self.pypi_config:
            self.generate_pip_conf()

    def generate_pypirc(self):
        pypirc_config = self.pypi_config['pypirc']

        if 'path' in pypirc_config:
            self.pypirc_path = pypirc_config.pop('path')

        print('generating .pypirc at %s' % self.pypirc_path)
        os.makedirs(os.path.dirname(self.pypirc_path), exist_ok=True)
        with open(self.pypirc_path, 'w') as pypirc:
            pypirc.write('[distutils]\n')
            pypirc.write('index-servers =\n')
            for server, config in pypirc_config['servers'].items():
                pypirc.write('    %s\n' % server)
            pypirc.write('\n')

            for server, config in pypirc_config['servers'].items():
                pypirc.write('[%s]\n' % server)
                for field in 'repository username password'.split():
                    if field in config:
                        pypirc.write('%s: %s\n' % (field, config[field]))
                pypirc.write('\n')

        print('.pypirc written to %s' % self.pypirc_path)

    def generate_pip_conf(self):
        pip_conf_config = self.pypi_config['pip.conf']

        if 'path' in pip_conf_config:
            self.pip_conf_path = pip_conf_config.pop('path')

        print('generating pip.conf at %s' % self.pip_conf_path)
        os.makedirs(os.path.dirname(self.pip_conf_path), exist_ok=True)
        with open(self.pip_conf_path, 'w') as pipconf:
            for config in pip_conf_config['entries']:
                pipconf.write('[%s]\n' % config['scope'])
                for key, value in config['args'].items():
                    pipconf.write('%s = %s\n' % (key, value))
                pipconf.write('\n')

        print('pip.conf written to %s' % self.pip_conf_path)

    def upload(self, server='pypi', path='dist/*'):
        self.run('twine upload --config-file=%s %s -r %s' % (self.pypirc_path, path, server))

    def if_package_versions_exist(self, package_directory=None, action=None):
        regex = re.compile('(.+)-(\d+\.\d+\.\d+\.?.*?)\..+')
        packages = {}

        for f in os.listdir(package_directory):
            if os.path.isfile(os.path.join(package_directory, f)):
                match = regex.search(f)
                package_name = match.group(1)
                version = match.group(2)

                if not package_name in packages:
                    packages[package_name] = []
                packages[package_name].append(version)

        print('found these packages at (%s): %s' % (package_directory, packages))

        existing_packages = []
        for package, versions in packages.items():
            existing_versions = self.get_existing_versions(package_name=package)
            intersection = set.intersection(set(existing_versions), set(versions))
            if bool(intersection):
                existing_packages.extend(['%s-%s' % (package, i) for i in intersection])

        if existing_packages and action:
            if action == 'error':
                print(existing_packages)
                raise ValueError('Package already exists! %s-%s' % (package_name, version))

    def get_existing_versions(self, **package_info):
        result = self.pip('install %s==' % package_info['name'])
        regex = re.compile('\(from versions: (.*)\)')
        match = regex.search(result['stdout'])

        versions = []
        if match:
            versions.extend(match.group(1).split(', '))

        print('existing package: %s' % package_info['name'])
        print('existing versions: %s' % versions)

        return versions

    def pip(self, command, hide_output=True):
        result = self.run('PIP_CONFIG_FILE=%s %s %s' % (self.pip_conf_path, os.getenv('PIP_EXE', 'pip'), command), check=False, hide_output=hide_output)
        return result



import os
from .service import Service as BaseService

class Service(BaseService):
    def __init__(self, config):
        super().__init__(config)
        self.registry = None

    def up(self):
        super().up()

        if 'steps' in self.config:
            self.run_steps(self.config['steps'])

    def run_steps(self, steps):
        for step in steps:
            getattr(self, step['type'])(**step.get('args', {}))

    def pull(self, image_name):
        self.run('docker pull %s' % image_name)

    def build(self, image=None, dockerfile=None, path='.', build_args={}):
        build_command = 'docker build -t %s' % image

        if dockerfile:
            build_command = '%s -f %s' % (build_command, dockerfile)

        if build_args:
            for key, value in build_args.items():
                build_command = '%s --build-arg %s=%s' % (build_command, key, value)

        build_command = '%s %s' % (build_command, path)

        self.run(build_command)

    def push(self, image=None, tags=[]):
        if 'latest' not in tags:
            tags.insert(0, 'latest')

        tags = [t.replace('/', '.') for t in tags]

        self.tag(image=image, tags=tags)

        for tag in tags:
            full_path = self.full_image_path(self.config['registry_path'], tag)
            self.run('docker push %s' % full_path)

    def tag(self, image=None, tags=[]):
        for tag in tags:
            full_path = self.full_image_path(self.config['registry_path'], tag)
            self.run('docker tag %s %s' % (image, full_path))

    def full_image_path(self, registry_path, tag):
        full_path = '%s:%s' % (registry_path, tag)
        if self.registry:
            full_path = '%s/%s' % (self.registry, full_path)
        return '"%s"' % full_path



from .gcloud_service import GCloudService

import os
import time
import yaml
import json

class Service(GCloudService):
    def up(self):
        super().up()

        self.cluster_name = self.config['cluster_name']

        if 'project' in self.config:
            self.run('gcloud config set project %s' % self.config['project'])
        if 'compute_zone' in self.config:
            self.run('gcloud config set compute/zone %s' % self.config['compute_zone'])

        self.ensure_cluster(self.cluster_name)

        if 'cluster_configs' in self.config:
            configs = self.config['cluster_configs']
            if not isinstance(configs, list):
                configs = [configs]

            for config in configs:
                self.apply(config)

        if 'wait_for_pending' in self.config:
            args = {}
            if isinstance(self.config['wait_for_pending'], dict):
                args = self.config['wait_for_pending']

            self.wait_for_pending_services(**args)

    def ensure_cluster(self, cluster_name=None):
        print('ensuring cluster: %s' % cluster_name)
        self.run('gcloud container clusters create %s || true' % cluster_name)
        self.run('gcloud container clusters get-credentials %s' % cluster_name)
        print('cluster ready: %s' % cluster_name)

    def apply(self, config):
        try:
            print('applying kube config:')
            print(json.dumps(config, indent=2, sort_keys=True, default=str))

            with open('file.yaml', 'w') as outfile:
                yaml.dump(config, outfile, default_flow_style=False)

            self.run('kubectl apply -f file.yaml')
        finally:
            os.remove('file.yaml')

    def wait_for_pending_services(self, **args):
        max_wait = args.get('max_wait', 300)
        wait_increment = args.get('wait_increment', 5)
        current_wait = 0

        status = '<pending>'
        while True:
            result = self.run('kubectl get service')
            if '<pending>' in result['stdout']:
                print('(%ss) found pending actions for cluster %s, waiting...' % (current_wait, self.cluster_name))
                time.sleep(wait_increment)
                current_wait += wait_increment
            else:
                break

            if current_wait > max_wait:
                print('timed out while waiting for pending actions, terminating')
                sys.exit(1)



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    author = 'Matt Groot',
    author_email = '',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: BSD License',
    ],
    description = ('CI/CD Automation Framework'),
    entry_points = {
        'console_scripts': [
            'servicer = servicer.servicer:main',
        ],
    },
    include_package_data = True,
    install_requires = [
        'PyYAML==3.12',
        'requests==2.18.4',
    ],
    keywords = 'ci cd automation environment service',
    license = 'BSD3',
    long_description = read('README.md'),
    name = 'servicer',
    packages = find_packages(),
    python_requires = '>=3.5',
    url = 'https://github.com/wmgroot/servicer',
    version = '0.7.12',
)
